import cv2
import numpy as np
from pathlib import Path

from util import *


def divide(mat, n0):
    """generate N"""
    ret = [n0]
    st, n = 0, mat.size
    while True:
        Ni = int(mat[st:st + ret[-1]].mean()) + ret[-1]
        st += ret[-1]
        if st + Ni >= n:
            ret.append(n - st)
            break
        ret.append(Ni)
    return ret


def gen_y(mat):
    """generate Y"""
    sub_n = divide(mat, N0)
    x = np.concatenate([delay_random(ni) for ni in sub_n])
    y = (x * 1e14).astype(np.int64) % 256
    return y.astype(np.int32)


def crossover(m, c0, z):
    """crossover diffusion"""
    m = m.reshape(2, -1).swapaxes(0, 1).reshape(-1)
    z = z.reshape(2, -1).swapaxes(0, 1).reshape(-1)
    m[0] ^= mp(c0, z[0])
    for i in range(1, m.size):
        m[i] ^= mp(m[i - 1], z[i])
    m[0] ^= mp(m[-1], z[0])
    for i in range(1, m.size):
        m[i] ^= mp(m[i - 1], z[i])
    m = m.reshape(-1, 2).swapaxes(0, 1).reshape(-1)
    return m


def encrypt(filename, gen_file=True, dst='data/cipher'):
    image = read(filename)
    h, w = image.shape
    mat = image.reshape(-1)
    n = mat.size
    z = random(n).argsort().astype(int)
    mat = mat[z]
    y = gen_y(mat)
    mat ^= y
    mat = crossover(mat, C0, z)
    mat[z] = mat.copy()
    cipher = mat.reshape(h, w)
    if gen_file:
        cv2.imwrite(f'{dst}/{Path(filename).stem}_e.bmp', cipher)
    return cipher


def de_crossover(m, c0, z):
    """remove crossover diffusion"""
    m = m.reshape(2, -1).swapaxes(0, 1).reshape(-1)
    z = z.reshape(2, -1).swapaxes(0, 1).reshape(-1)
    for i in reversed(range(1, m.size)):
        m[i] ^= mp(m[i - 1], z[i])
    m[0] ^= mp(m[-1], z[0])
    for i in reversed(range(1, m.size)):
        m[i] ^= mp(m[i - 1], z[i])
    m[0] ^= mp(c0, z[0])
    m = m.reshape(-1, 2).swapaxes(0, 1).reshape(-1)
    return m


def de_divide(m, y, n0):
    """remove confusion"""
    now, nk = 0, n0
    while True:
        if now + nk >= m.size:
            m[now:] ^= y[:m.size - now]
            break
        m[now:now + nk] ^= y[:nk]
        increment = int(m[now:now + nk].mean()) % 256
        now += nk
        nk += increment
    return m


def solve(filename, n0, c_in, dst=None):
    """decrypt cipher-image with the obtained key"""
    y = np.loadtxt('data/y.txt', dtype=int)
    z = np.loadtxt('data/z.txt', dtype=int)
    cipher = read(filename)
    h, w = cipher.shape
    mat = cipher.reshape(-1)
    mat = mat[z]
    mat = de_crossover(mat, c_in, z)
    cv2.imwrite('data/decipher-p/de_crossover.bmp', mat.reshape(h, w))
    mat = de_divide(mat, y, n0)
    cv2.imwrite('data/decipher-p/de_divide.bmp', mat.reshape(h, w))
    mat[z] = mat.copy()
    plain = mat.reshape(h, w)
    cv2.imwrite(f'data/decipher-p/decipher.bmp', plain)
    cv2.imwrite(f'{dst or "data/decipher"}/{Path(filename).stem}_d.bmp', plain)


def batch_encrypt(ori, dst):
    for file in Path(ori).iterdir():
        cipher = encrypt(str(file), False)
        cv2.imwrite(f'{dst}/{file.stem}.bmp', cipher)


if __name__ == '__main__':
    encrypt('data/gen/p/b0.bmp')
    solve('data/cipher/p10_e.bmp', N0, C0)
