import numpy as np
import cv2


N0 = 100
C0 = 216
np.random.seed(834263081)
X0 = np.random.sample(50) * 3


def random(n, q=0.75, beta=3.7488464):
    """Logistic map"""
    ret = np.zeros(n, float)
    for i in range(n):
        q = ret[i] = beta * q * (1 - q)
    return ret


def delay_random(n, x0=X0, alpha=6, m=19.5, h=0.1):
    """discretized Ikeda map"""
    t = x0.size
    ret = np.zeros(n + t, float)
    ret[:t] = x0
    for i in range(t, t + n):
        ret[i] = ret[i-1] + (-alpha * ret[i-1] + m * np.sin(ret[i - t])) * h
    return ret[t:]


def read(file):
    """read grayscale image"""
    return cv2.imread(file, 0).astype(int) if isinstance(file, str) else file


def feed(file):
    """flatten"""
    return read(file).reshape(-1)


def mp(a, b):
    """modulo plus"""
    return (a + b) % 256


def ms(a, b):
    """modulo subtraction"""
    return ((a - b) % 256 + 256) % 256


def batch_read(prefix, num):
    """stack images"""
    return np.stack([read(f'{prefix}{i}.bmp').reshape(-1) for i in range(num)],
                    axis=1)


def generate_pair(index, num):
    """generate I1 corresponding to I0"""
    for i in range(num):
        image = read(f'data/plain/p{i}.bmp')
        cv2.imwrite(f'data/gen/p/a{i}.bmp', image)
        delta = 1 if image.flat[index] != 255 else -1
        image.flat[index] += delta
        cv2.imwrite(f'data/gen/p/b{i}.bmp', image)
